@@ -8,12 +8,11 @@ import confuse
 import json
 import logging
 from .message_mapping import (
-    get_vast_intel_type,
-    get_ioc,
-    matcher_result_to_threatbus_sighting,
-    query_result_to_threatbus_sighting,
-    to_vast_ioc,
-    to_vast_query,
+    get_vast_type_and_value,
+    indicator_to_vast_matcher_ioc,
+    indicator_to_vast_query,
+    matcher_result_to_sighting,
+    query_result_to_sighting,
 )
 from pyvast import VAST
 import random
@@ -22,8 +21,11 @@ import socket
 from string import ascii_lowercase as letters
 import sys
 from .metrics import Gauge, Summary
+from stix2 import parse, Indicator, Sighting
 from threatbus.logger import setup as setup_logging_threatbus
-from threatbus.data import Intel, IntelDecoder, Sighting, SightingEncoder, Operation
+from threatbus.data import Operation, ThreatBusSTIX2Constants
+from threatbus.stix2_helpers import split_object_path_and_value
+
 import time
 import zmq
 
@@ -80,9 +82,13 @@ def validate_config(config: confuse.Subview):
     config["vast_binary"].get(str)
     config["threatbus"].get(str)
     config["snapshot"].get(int)
-    config["retro_match"].get(bool)
+    live_match = config["live_match"].get(bool)
+    retro_match = config["retro_match"].get(bool)
+    assert (
+        live_match or retro_match
+    ), "please enable at least one kind of matching (live_match and/or retro_match)"
     config["retro_match_max_events"].get(int)
-    config["unflatten"].get(bool)
+    config["retro_match_timeout"].get(float)
     config["max_background_tasks"].get(int)
 
     # fallback values for the optional arguments
@@ -117,9 +123,10 @@ async def start(
     vast_endpoint: str,
     zmq_endpoint: str,
     snapshot: int,
+    live_match: bool,
     retro_match: bool,
     retro_match_max_events: int,
-    unflatten: bool,
+    retro_match_timeout: float,
     max_open_files: int,
     metrics_interval: int,
     metrics_filename: str,
@@ -128,15 +135,16 @@ async def start(
 ):
     """
     Starts the app between the two given endpoints. Subscribes the configured
-    VAST instance for threat intelligence (IoCs) and reports new intelligence to
+    VAST instance for threat intelligence (IoCs) and reports new sightings to
     Threat Bus.
     @param cmd The vast binary command to use with PyVAST
     @param vast_endpoint The endpoint of a running VAST node ('host:port')
     @param zmq_endpoint The ZMQ management endpoint of Threat Bus ('host:port')
-    @param snapshot An integer value to request n days of past intel items
-    @param retro_match Boolean flag to use retro-matching over live-matching
+    @param snapshot An integer value to request n days of historical IoC items
+    @param live_match Boolean flag to enable live-matching
+    @param retro_match Boolean flag to enable retro-matching
     @param retro_match_max_events Max amount of retro match results
-    @param unflatten Boolean flag to unflatten JSON when received from VAST
+    @param retro_match_timeout Interval after which to terminate the retro-query
     @param max_open_files The maximum number of concurrent background tasks for VAST queries.
     @param merics_interval The interval in seconds to bucketize metrics
     @param metrics_filename The filename (system path) to store metrics at
@@ -151,7 +159,7 @@ async def start(
     assert await vast.test_connection() is True, "Cannot connect to VAST"
 
     logger.debug(f"Calling Threat Bus management endpoint {zmq_endpoint}")
-    reply = subscribe(zmq_endpoint, "threatbus/intel", snapshot)
+    reply = subscribe(zmq_endpoint, "stix2/indicator", snapshot)
     if not reply_is_success(reply):
         logger.error("Subscription failed")
         return
@@ -175,7 +183,7 @@ async def start(
         asyncio.create_task(heartbeat(zmq_endpoint, p2p_topic, interval=5))
     )
 
-    intel_queue = asyncio.Queue()
+    indicator_queue = asyncio.Queue()
     sightings_queue = asyncio.Queue()
     async_tasks.append(
         asyncio.create_task(
@@ -184,7 +192,7 @@ async def start(
     )
 
     async_tasks.append(
-        asyncio.create_task(receive(pub_endpoint, p2p_topic, intel_queue))
+        asyncio.create_task(receive(pub_endpoint, p2p_topic, indicator_queue))
     )
 
     async_tasks.append(
@@ -192,11 +200,12 @@ async def start(
             match_intel(
                 vast_binary,
                 vast_endpoint,
-                intel_queue,
+                indicator_queue,
                 sightings_queue,
+                live_match,
                 retro_match,
                 retro_match_max_events,
-                unflatten,
+                retro_match_timeout,
             )
         )
     )
@@ -204,7 +213,7 @@ async def start(
     if retro_match:
         # add metrics for retro-matching to the metric output
         metrics += [s_retro_matches_per_ioc, s_retro_query_time_s_per_ioc]
-    else:
+    if live_match:
         # add metrics for live-matching to the metric output
         metrics.append(g_live_matcher_sightings)
         async_tasks.append(
@@ -251,14 +260,14 @@ async def write_metrics(every: int, to: str):
         await asyncio.sleep(every)
 
 
-async def receive(pub_endpoint: str, topic: str, intel_queue: asyncio.Queue):
+async def receive(pub_endpoint: str, topic: str, indicator_queue: asyncio.Queue):
     """
     Starts a zmq subscriber on the given endpoint and listens for new messages
     that are published on the given topic (zmq prefix matching). Depending on
-    the topic suffix, Intel items (IoCs) are enqueued to the intel_queue.
+    the topic suffix, Indicators are enqueued to the indicator_queue.
     @param pub_endpoint A host:port string to connect to via zmq
     @param topic The topic prefix to subscribe to intelligence items
-    @param intel_queue The queue to put arriving IoCs into
+    @param indicator_queue The queue to put arriving IoCs into
     """
     global logger
     socket = zmq.Context().socket(zmq.SUB)
@@ -278,128 +287,161 @@ async def receive(pub_endpoint: str, topic: str, intel_queue: asyncio.Queue):
                 logger.error(f"Error decoding message: {e}")
                 continue
             # the topic is suffixed with the message type
-            if not topic.endswith("intel"):
+            if not topic.endswith("indicator"):
                 # pyvast-threatbus is not (yet) interested in Sightings or SnapshotRequests
                 logger.debug(f"Skipping unsupported message: {msg}")
                 continue
-            await intel_queue.put(msg)
+            await indicator_queue.put(msg)
         else:
             await asyncio.sleep(0.05)  # free event loop for other tasks
 
 
 async def retro_match_vast(
-    vast_binary,
-    vast_endpoint,
-    retro_match_max_events,
-    intel,
-    sightings_queue,
-    unflatten,
+    vast_binary: str,
+    vast_endpoint: str,
+    retro_match_max_events: int,
+    retro_match_timeout: float,
+    indicator: Indicator,
+    sightings_queue: asyncio.Queue,
 ):
     """
-    Turns the given intel into a valid VAST query and forwards all all query
-    results (sightings) to the sightings_queue.
+    Turns the given STIX-2 Indicator into a valid VAST query and forwards all
+    query results (sightings) to the sightings_queue.
     @param vast_binary The vast binary command to use with PyVAST
     @param vast_endpoint The endpoint of a running vast node ('host:port')
     @param retro_match_max_events  Max amount of retro match results
-    @param intel The IoC to query VAST for
+    @param retro_match_timeout Interval after which to terminate the retro-query
+    @param indicator The STIX-2 Indicator to query VAST for
     @param sightings_queue The queue to put new sightings into
-    @param unflatten Boolean flag to unflatten JSON when received from VAST
     """
     start = time.time()
-    query = to_vast_query(intel)
+    query = indicator_to_vast_query(indicator)
     if not query:
         return
     global logger, max_open_tasks
     async with max_open_tasks:
         vast = VAST(binary=vast_binary, endpoint=vast_endpoint, logger=logger)
-        proc = await vast.export(max_events=retro_match_max_events).json(query).exec()
+        kwargs = {}
+        if retro_match_max_events > 0:
+            kwargs["max_events"] = retro_match_max_events
+        proc = await vast.export(**kwargs).json(query).exec()
+        retro_result = None
+        try:
+            retro_result = await asyncio.wait_for(
+                proc.communicate(),
+                timeout=retro_match_timeout if retro_match_timeout > 0 else None,
+            )
+        except asyncio.TimeoutError:
+            proc.terminate()
+            logger.error(
+                f"Timeout after {retro_match_timeout}s in retro-query for indicator {indicator}"
+            )
+        if not retro_result or len(retro_result) != 2:
+            return
         reported = 0
-        while not proc.stdout.at_eof():
-            line = (await proc.stdout.readline()).decode().rstrip()
+        stdout = retro_result[0]
+        for line in stdout.decode().split("\n"):
+            line = line.rstrip()
             if line:
-                sighting = query_result_to_threatbus_sighting(line, intel, unflatten)
+                sighting = query_result_to_sighting(line, indicator)
                 if not sighting:
                     logger.error(f"Could not parse VAST query result: {line}")
                     continue
                 reported += 1
                 await sightings_queue.put(sighting)
-        logger.debug(f"Retro-matched {reported} sighting(s) for intel: {intel}")
+        logger.debug(f"Retro-matched {reported} sighting(s) for indicator: {indicator}")
         s_retro_matches_per_ioc.observe(reported)
         s_retro_query_time_s_per_ioc.observe(time.time() - start)
 
 
-async def ingest_vast_ioc(vast_binary, vast_endpoint, intel):
+async def ingest_vast_ioc(vast_binary: str, vast_endpoint: str, indicator: Indicator):
     """
-    Ingests the given intel as IoC into a VAST matcher.
+    Converts the given STIX-2 Indicator to a VAST-compatible IoC and ingests it
+    via a VAST matcher.
     @param vast_binary The vast binary command to use with PyVAST
     @param vast_endpoint The endpoint of a running vast node ('host:port')
-    @param intel The IoC to query VAST for
+    @param indicator The STIX-2 Indicator to query VAST for
     """
     global logger
-    ioc = to_vast_ioc(intel)
-    if not ioc:
-        logger.error(f"Unable to convert Intel to VAST compatible IoC: {intel}")
-        return
-    vast = VAST(binary=vast_binary, endpoint=vast_endpoint, logger=logger)
-    proc = await vast.import_().json(type="intel.indicator").exec(stdin=ioc)
-    await proc.wait()
-    logger.debug(f"Ingested intel for live matching: {intel}")
-
-
-async def remove_vast_ioc(vast_binary, vast_endpoint, intel):
-    """
-    Removes the given intel as IoC from a VAST matcher.
-    @param vast_binary The vast binary command to use with PyVAST
-    @param vast_endpoint The endpoint of a running vast node ('host:port')
-    @param intel The IoC to query VAST for
-    """
-    global logger, matcher_name
-    intel_type = get_vast_intel_type(intel)
-    ioc = get_ioc(intel)
-    if not ioc or not intel_type:
+    vast_ioc = indicator_to_vast_matcher_ioc(indicator)
+    if not vast_ioc:
         logger.error(
-            f"Cannot remove intel with missing intel_type or indicator: {intel}"
+            f"Unable to convert STIX-2 Indicator to VAST compatible IoC. Is it a point IoC? {indicator}"
         )
         return
     vast = VAST(binary=vast_binary, endpoint=vast_endpoint, logger=logger)
-    await vast.matcher().ioc_remove(matcher_name, ioc, intel_type).exec()
-    logger.debug(f"Removed indicator {intel}")
+    proc = await vast.import_(type="intel.indicator").json().exec(stdin=vast_ioc)
+    await proc.wait()
+    logger.debug(f"Ingested indicator for VAST live matching: {indicator}")
+
+
+async def remove_vast_ioc(vast_binary: str, vast_endpoint: str, indicator: Indicator):
+    """
+    Converts the given STIX-2 Indicator to a VAST-compatible IoC and removes it
+    from the VAST matcher.
+    @param vast_binary The vast binary command to use with PyVAST
+    @param vast_endpoint The endpoint of a running vast node ('host:port')
+    @param indicator The STIX-2 Indicator to remove from VAST
+    """
+    global logger, matcher_name
+    type_and_value = get_vast_type_and_value(indicator.pattern)
+    if not type_and_value:
+        logger.debug(f"Cannot remove IoC from VAST. Is it a point IoC? {indicator}")
+        return None
+    (vast_type, ioc_value) = type_and_value
+    vast = VAST(binary=vast_binary, endpoint=vast_endpoint, logger=logger)
+    # TODO pass matcher_name once VAST supports more fine-grained deletion
+    await vast.matcher().intel().remove(ioc_value, vast_type).exec()
+    logger.debug(f"Removed indicator from VAST live matching: {indicator}")
 
 
 async def match_intel(
     vast_binary: str,
     vast_endpoint: str,
-    intel_queue: asyncio.Queue,
+    indicator_queue: asyncio.Queue,
     sightings_queue: asyncio.Queue,
+    live_match: bool,
     retro_match: bool,
     retro_match_max_events: int,
-    unflatten: bool,
+    retro_match_timeout: float,
 ):
     """
-    Reads from the intel_queue and matches all IoCs, either via VAST's
+    Reads from the indicator_queue and matches all IoCs, either via VAST's
     live-matching or retro-matching.
     @param vast_binary The vast binary command to use with PyVAST
     @param vast_endpoint The endpoint of a running vast node ('host:port')
-    @param intel_queue The queue to read new IoCs from
+    @param indicator_queue The queue to read new IoCs from
     @param sightings_queue The queue to put new sightings into
-    @param retro_match Boolean flag to use retro-matching over live-matching
+    @param live_match Boolean flag to use retro-matching
+    @param retro_match Boolean flag to use live-matching
     @param retro_match_max_events  Max amount of retro match results
-    @param unflatten Boolean flag to unflatten JSON when received from VAST
+    @param retro_match_timeout Interval after which to terminate the retro-query
     """
     global logger, open_tasks
     while True:
-        msg = await intel_queue.get()
+        msg = await indicator_queue.get()
         try:
-            intel = json.loads(msg, cls=IntelDecoder)
+            indicator = parse(msg, allow_custom=True)
         except Exception as e:
-            logger.warning(f"Failed to decode intel item {msg}: {e}")
+            logger.warning(f"Failed to decode STIX-2 Indicator item {msg}: {e}")
             continue
-        if type(intel) is not Intel:
+        if type(indicator) is not Indicator:
             logger.warning(
-                f"Ignoring unknown message type, expected Intel: {type(intel)}"
+                f"Ignoring unknown message type, expected STIX-2 Indicator: {type(indicator)}"
             )
             continue
-        if intel.operation == Operation.ADD:
+        if (
+            ThreatBusSTIX2Constants.X_THREATBUS_UPDATE.value
+            in indicator.object_properties()
+            and indicator.x_threatbus_update == Operation.REMOVE.value
+        ):
+            g_iocs_removed.inc()
+            if live_match:
+                asyncio.create_task(
+                    remove_vast_ioc(vast_binary, vast_endpoint, indicator)
+                )
+        else:
+            # add new Indicator to matcher / query Indicator retrospectively
             g_iocs_added.inc()
             if retro_match:
                 asyncio.create_task(
@@ -407,21 +449,16 @@ async def match_intel(
                         vast_binary,
                         vast_endpoint,
                         retro_match_max_events,
-                        intel,
+                        retro_match_timeout,
+                        indicator,
                         sightings_queue,
-                        unflatten,
                     )
                 )
-            else:
-                asyncio.create_task(ingest_vast_ioc(vast_binary, vast_endpoint, intel))
-        elif intel.operation == Operation.REMOVE:
-            g_iocs_removed.inc()
-            if retro_match:
-                continue
-            asyncio.create_task(remove_vast_ioc(vast_binary, vast_endpoint, intel))
-        else:
-            logger.warning(f"Unsupported operation for indicator: {intel}")
-        intel_queue.task_done()
+            if live_match:
+                asyncio.create_task(
+                    ingest_vast_ioc(vast_binary, vast_endpoint, indicator)
+                )
+        indicator_queue.task_done()
 
 
 async def live_match_vast(
@@ -439,7 +476,8 @@ async def live_match_vast(
     vast = VAST(binary=vast_binary, endpoint=vast_endpoint, logger=logger)
     matcher_name = "threatbus-" + "".join(random.choice(letters) for i in range(10))
     proc = await vast.matcher().start(name=matcher_name).exec()
-    while True:
+    # returncode is None as long as the process did not terminate yet
+    while proc.returncode is None:
         data = await proc.stdout.readline()
         if not data:
             if not await vast.test_connection():
@@ -447,12 +485,18 @@ async def live_match_vast(
                 # TODO reconnect
             continue
         vast_sighting = data.decode("utf-8").rstrip()
-        sighting = matcher_result_to_threatbus_sighting(vast_sighting)
+        sighting = matcher_result_to_sighting(vast_sighting)
         if not sighting:
             logger.error(f"Cannot parse sighting-output from VAST: {vast_sighting}")
             continue
         g_live_matcher_sightings.inc()
         await sightings_queue.put(sighting)
+    stderr = await proc.stderr.read()
+    if stderr:
+        logger.error(
+            "VAST matcher process exited with message: {}".format(stderr.decode())
+        )
+    logger.critical("Unexpected exit of VAST matcher process.")
 
 
 async def invoke_cmd_for_context(cmd: str, context: dict, ioc: str = "%ioc"):
@@ -499,8 +543,8 @@ async def report_sightings(
     @param transform_cmd The command to use to pipe sightings to. Treated
         as template string: occurrences of '%ioc' in the cmd string get replaced
         with the matched IoC.
-    @param report_data If True, only report sighting.context.data instead of the
-        whole sighting.
+    @param report_data If True, only report context data of the sighting instead
+        of the whole thing.
     """
     global logger
     if transform_cmd:
@@ -512,7 +556,7 @@ async def report_sightings(
     else:
         socket = zmq.Context().socket(zmq.PUB)
         socket.connect(f"tcp://{sub_endpoint}")
-        topic = "threatbus/sighting"
+        topic = "stix2/sighting"
         logger.info(f"Forwarding sightings to Threat Bus at {sub_endpoint}/{topic}")
     while True:
         sighting = await sightings_queue.get()
@@ -521,33 +565,91 @@ async def report_sightings(
                 f"Ignoring unknown message type, expected Sighting: {type(sighting)}"
             )
             continue
-        if transform_cmd and sighting.context:
-            ioc = sighting.ioc
-            if sighting.ioc and type(sighting.ioc) is tuple:
-                ioc = sighting.ioc[0]  # use first value of tuple
-            context_str = await invoke_cmd_for_context(
-                transform_cmd, sighting.context, ioc
+        if transform_cmd:
+            sighting = await transform_context(sighting, transform_cmd)
+        if sink:
+            context = (
+                sighting.x_threatbus_sighting_context
+                if ThreatBusSTIX2Constants.X_THREATBUS_SIGHTING_CONTEXT.value
+                in sighting.object_properties()
+                else None
             )
-            try:
-                context = json.loads(context_str)
-                sighting.context = context
-            except Exception as e:
-                logger.error(
-                    f"Cannot parse transformed sighting context (expecting JSON): {context_str}",
-                    e,
+            if not context:
+                logger.warn(
+                    f"Cannot report sighting context to custom sink because no context data is found in the sighting {sighting}"
                 )
                 continue
-        if sink:
             if sink.lower() == "stdout":
-                print(json.dumps(sighting.context))
+                print(json.dumps(context))
             else:
-                await invoke_cmd_for_context(sink, sighting.context)
+                await invoke_cmd_for_context(sink, context)
         else:
-            socket.send_string(f"{topic} {json.dumps(sighting, cls=SightingEncoder)}")
+            socket.send_string(f"{topic} {sighting.serialize()}")
         sightings_queue.task_done()
         logger.debug(f"Reported sighting: {sighting}")
 
 
+async def transform_context(sighting: Sighting, transform_cmd: str) -> Sighting:
+    """
+    Transforms the context of a sighting using the command configured in
+    `transform_context`
+    @param sighting the sighting as it was reported by VAST
+    @param transform_cmd The command to use to pipe sightings to. Treated
+        as template string: occurrences of '%ioc' in the cmd string get replaced
+        with the matched IoC.
+    @return a copy of the original sighting with the x_threatbus_context field
+        set and transformed accordingly
+    """
+    context = (
+        sighting.x_threatbus_sighting_context
+        if ThreatBusSTIX2Constants.X_THREATBUS_SIGHTING_CONTEXT.value
+        in sighting.object_properties()
+        else None
+    )
+    if not context:
+        logger.error(
+            f"Cannot invoke `transform_context` command because no context data is found in the sighting {sighting}"
+        )
+        return
+    indicator = (
+        sighting.x_threatbus_indicator
+        if ThreatBusSTIX2Constants.X_THREATBUS_INDICATOR.value
+        in sighting.object_properties()
+        else None
+    )
+    if indicator:
+        _, ioc_value = split_object_path_and_value(indicator.pattern)
+    else:
+        # try to find the indicator value instead
+        ioc_value = (
+            sighting.x_threatbus_indicator_value
+            if ThreatBusSTIX2Constants.X_THREATBUS_INDICATOR_VALUE.value
+            in sighting.object_properties()
+            else None
+        )
+    if not ioc_value:
+        logger.error(
+            f"Cannot invoke `transform_context` command because no indicator value is found in the sighting {sighting}"
+        )
+        return
+    transformed_context_raw = await invoke_cmd_for_context(
+        transform_cmd, context, ioc_value
+    )
+    try:
+        transformed_context = json.loads(transformed_context_raw)
+        # recreate the sighting with the new transformed context
+        ser = json.loads(sighting.serialize())
+        ser[
+            ThreatBusSTIX2Constants.X_THREATBUS_SIGHTING_CONTEXT.value
+        ] = transformed_context
+        return parse(json.dumps(ser), allow_custom=True)
+    except Exception as e:
+        logger.error(
+            f"Cannot parse transformed sighting context (expecting JSON): {transformed_context_raw}",
+            e,
+        )
+
+
 def send_manage_message(endpoint: str, action: dict, timeout: int = 5):
     """
     Sends a 'management' message, following the threatbus-zmq-app protocol to
@@ -589,11 +691,11 @@ def reply_is_success(reply: dict):
 
 def subscribe(endpoint: str, topic: str, snapshot: int, timeout: int = 5):
     """
-    Subscribes this app to the Threat Bus for the given topic. Requests an
-    optional snapshot of past intelligence data.
+    Subscribes this app to Threat Bus for the given topic. Requests an optional
+    snapshot of historical indicators.
     @param endpoint The ZMQ management endpoint of Threat Bus ('host:port')
     @param topic The topic to subscribe to
-    @param snapshot An integer value to request n days of past intel items
+    @param snapshot An integer value to request n days of historical IoC items
     @param timeout The period after which the connection attempt is aborted
     """
     global logger
@@ -674,9 +776,10 @@ def main():
                     config["vast"].get(),
                     config["threatbus"].get(),
                     config["snapshot"].get(),
+                    config["live_match"].get(),
                     config["retro_match"].get(),
                     config["retro_match_max_events"].get(),
-                    config["unflatten"].get(),
+                    config["retro_match_timeout"].get(),
                     config["max_background_tasks"].get(),
                     config["metrics"]["interval"].get(),
                     config["metrics"]["filename"].get(),
